{

AiChip_Synth_003.spin
=====================

For an overview description see the AiChip_Synth_001.spin file.
                                                          
This is essentially a rewrite of AiChip_Synth_002.spin with the vcoParams[] array now named vco1[]
and accessed as a global long[] array via pointers. This will allow multiple VCO's to be used in
future versions.

TIMEOUT_TICKS in SynthPlayer has been increased again to take into account the increased exeution
time.

}

CON

  _CLKMODE         = XTAL1 + PLL16x                     ' 80MHz System Clock Speed
  _XINFREQ         = 5_000_000

OBJ

  tv               : "TV_Text"                          ' Debugging only - can be deleted
  
CON

  PIN_LEFT         = 10                                 ' Set -Ve if not used
  PIN_RIGHT        = 11                                 ' Set -Ve if not used
                                                                                
  PIN_TV           = 12                                 ' Base TV Pin

  NONE             = 0                                  ' Wave shapes for VCO
  SINE             = 1
  TRIANGLE         = 2
  SQUARE           = 3
  SAW_UP           = 4
  SAW_DOWN         = 5
  RAMP_UP          = 6
  RAMP_DOWN        = 7
  SPIKE            = 8
  SPIKE_UP         = 9
  SPIKE_DOWN       = 10
  NOISE            = 11
  
  VCO_RESET        = 0 * 4                              ' VCO control parameters
  VCO_FREQ         = 1 * 4
  VCO_FREQSTEPRATE = 2 * 4
  VCO_WAVETYPE     = 3 * 4
  VCO_TABLEINDEX   = 4 * 4
  VCO_OUTPUTLEVEL  = 5 * 4

  VCO_PARAM_MAX    = 5 * 4
  
VAR

  long vco1[ ( VCO_PARAM_MAX / 4 ) + 1 ]
  
PUB Synth | freq

  tv.start( PIN_TV )

  StartSynthPlayerInOwnCog

  long[ @vco1 + VCO_WAVETYPE ] := SINE
  repeat
    repeat long[ @vco1 + VCO_FREQ ] from 440 to 4_000
      pauseMs(10)
      tv.out($01)
      tv.hex(long[ @vco1 + VCO_FREQSTEPRATE ],8)
      tv.out(" ")
      tv.dec(long[ @vco1 + VCO_FREQ ])
      tv.str(String("Hz  "))
    repeat long[ @vco1 + VCO_FREQ ] from 4_000 to 440
      pauseMs(10)
      tv.out($01)
      tv.hex(long[ @vco1 + VCO_FREQSTEPRATE ],8)
      tv.out(" ")
      tv.dec(long[ @vco1 + VCO_FREQ ])
      tv.str(String("Hz  "))

PRI PauseMs( mS )

  waitcnt( CLKFREQ / 1000 * mS + CNT )

' ******************************************************************************************************

CON

  TIMEOUT_TICKS = 2_000

VAR

  long freqAtOneStepPerTick
  long noiseSource
  
  long stackSynthPlayer[50]                             ' Guessed size

PUB StartSynthPlayerInOwnCog

  cognew( SynthPlayer, @stackSynthPlayer )
  
  pauseMs(100) ' Give the SynthPlayer time to get loaded and running

PRI SynthPlayer

  if ( PIN_LEFT => 0 ) and ( PIN_LEFT =< 31 )
    DIRA[ PIN_LEFT  ] := 1

  if ( PIN_RIGHT => 0 ) and ( PIN_RIGHT =< 31 ) and ( PIN_LEFT <> PIN_RIGHT)
    DIRA[ PIN_RIGHT ] := 1

  CTRA := %00110 << 26 + 0 << 9 + PIN_LEFT
  CTRB := %00110 << 26 + 0 << 9 + PIN_RIGHT

  freqAtOneStepPerTick := CLKFREQ / ( TIMEOUT_TICKS * 256 )

  noiseSource := CNT
  
  repeat
    Vco( @vco1 )                                        ' The main synthesiser function
    Dac( @vco1 )
    waitcnt( CNT + TIMEOUT_TICKS )                      ' Very poor quality timing loop
 
PRI Vco( vcoPtr ) | vcoTableIndex

  long[ vcoPtr + VCO_FREQSTEPRATE ] := $01_00_00_00 / freqAtOneStepPerTick * long[ vcoPtr + VCO_FREQ ]
 
  if long[ vcoPtr + VCO_RESET ]
    long[ vcoPtr + VCO_RESET ] := 0
    vcoTableIndex := 0
  else
    vcoTableIndex := long[ vcoPtr + VCO_TABLEINDEX ]
  
  case long[ vcoPtr + VCO_WAVETYPE ]
    NONE :
      result := $80_00_00_00
    SINE :
      result := ( byte[ @SineTable + ( vcoTableIndex >> 24 ) ] ) << 24
    TRIANGLE :               
      if vcoTableIndex => 0
        result := ( vcoTableIndex << 1 )
      else
        result := ( vcoTableIndex << 1 ) ^ $FF_FF_FF_FF
    SQUARE :
      result := vcoTableIndex < 0
    SAW_UP :
      result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) + $40 ) ] ^ $FF ) << 24
    SAW_DOWN :
      result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) + $40 ) ]       ) << 24
    RAMP_UP :
      result := vcoTableIndex
    RAMP_DOWN :
      result := vcoTableIndex ^ $FF_FF_FF_FF
    SPIKE :
      case vcoTableIndex >> 30
        %00 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 ) ^ $80 ) ] | $80 ) << 24
        %01 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 )       ) ] & $7F ) << 24
        %10 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 ) ^ $80 ) ] & $7F ) << 24
        %11 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 )       ) ] | $80 ) << 24
    SPIKE_UP :
      if vcoTableIndex => 0
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 )       ) ] ^ $80 ) << 24
      else
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) | $C0 ) ] ^ $80 ) << 24
    SPIKE_DOWN :
      if vcoTableIndex => 0
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 )       ) ] ^ $7F ) << 24
      else
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) | $C0 ) ] ^ $7F ) << 24
    NOISE :
      ? noiseSource                                       
      result := ( vcoTableIndex  & $80_00_00_00 ) | ( noiseSource & $7F_FF_FF_FF )

  long[ vcoPtr + VCO_TABLEINDEX ] := vcoTableIndex + long[ vcoPtr + VCO_FREQSTEPRATE ]
                
  long[ vcoPtr + VCO_OUTPUTLEVEL ] := result
    
PRI Dac( vcoPtr )

  FRQA := long[ vcoPtr + VCO_OUTPUTLEVEL ]
  FRQB := FRQA

DAT

SineTable     byte 128,131,134,137,140,143,146,149,152,156,159,162,165,168,171,174
              byte 176,179,182,185,188,191,193,196,199,201,204,206,209,211,213,216
              byte 218,220,222,224,226,228,230,232,234,236,237,239,240,242,243,245
              byte 246,247,248,249,250,251,252,252,253,254,254,255,255,255,255,255
              byte 255,255,255,255,255,255,254,254,253,252,252,251,250,249,248,247
              byte 246,245,243,242,240,239,237,236,234,232,230,228,226,224,222,220
              byte 218,216,213,211,209,206,204,201,199,196,193,191,188,185,182,179
              byte 176,174,171,168,165,162,159,156,152,149,146,143,140,137,134,131
              byte 128,124,121,118,115,112,109,106,103, 99, 96, 93, 90, 87, 84, 81 
              byte  79, 76, 73, 70, 67, 64, 62, 59, 56, 54, 51, 49, 46, 44, 42, 39 
              byte  37, 35, 33, 31, 29, 27, 25, 23, 21, 19, 18, 16, 15, 13, 12, 10 
              byte   9,  8,  7,  6,  5,  4,  3,  3,  2,  1,  1,  0,  0,  0,  0,  0  
              byte   0,  0,  0,  0,  0,  0,  1,  1,  2,  3,  3,  4,  5,  6,  7,  8  
              byte   9, 10, 12, 13, 15, 16, 18, 19, 21, 23, 25, 27, 29, 31, 33, 35 
              byte  37, 39, 42, 44, 46, 49, 51, 54, 56, 59, 62, 64, 67, 70, 73, 76 
              byte  79, 81, 84, 87, 90, 93, 96, 99,103,106,109,112,115,118,121,124
          