{

AiChip_Synth_004.spin
=====================

For an overview description see the AiChip_Synth_001.spin file.
                                                          
This is an extension of AiChip_Synth_003.spin with Amplitude Modulation added to the VCO. The
VCO_AMPTR paramter points to a long whose MSB contains a value ($00-$FF=0%-100%) which sets the
output level. When VCO_AMPTR is set to zero no Amplitude Modulation is applied. VCO_AMDEPTH is
a byte value which determines how much modulation ($00-$FF,None-Full) should be applied, This
allows the VCO to be varied from being gently 'warbled' to being switched on and off.

Amplitude Modulation is a litle complicated because we have a $0000_0000-$FFFF_FFFF value which
represents -N to +N with 0 represented by $8000_0000. Thus Amplitude Modulation has to compress
the signal with respect to $8000_0000, ie, A swing from $0000_0000 to $FFFF_FFFF when modulated
to 50% becomes a swing between $4000_0000 to $BFFF_FFFF, *not* $0000_0000 to $7FFF_FFFF.

I'm not sure if this is actually necessary, but it seemed to be a sensible ting to do.

Amplitude Depth is also a little complicted. When set to 50% the overall Amplitude Modulation is
set to 50% to 100% of full volume, when set to 25% it is 75% to 100% of full volume.

Ideally all byte multiplications should be done with (A*B)/255 to get a 0-255 result ( where A
and B are 0-255 ), but we use (A/256)*B for speed and disregard the slight imperfection.

NOTE : The VCO_OUTPUTLEVEL parameter has moved up the table. This is so that Amplitude Modulation
can be done by pointing at a VCO or an ADSR ( when added ). All entities which can modulate will
have their output levels placed in the first parameter position so the VCO doesn't care what it
is pointed at. Being the first parameter also means that a simple long holding a pot value for
example can be used as a modulation source.
 
This demo code includes three VCO's. One running at 1kHz and another (LFO) running at 1Hz to
Amplitude Modulate that. The loop in SynthPlayer has to call the Vco() routine twice to handle
both oscillators.

TIMEOUT_TICKS in SynthPlayer has been increased again to take into account the increased exeution
time.

}

CON

  _CLKMODE         = XTAL1 + PLL16x                     ' 80MHz System Clock Speed
  _XINFREQ         = 5_000_000

OBJ

  tv               : "TV_Text"                          ' Debugging only - can be deleted
  
CON

  PIN_LEFT         = 10                                 ' Set -Ve if not used
  PIN_RIGHT        = 11                                 ' Set -Ve if not used
                                                                                
  PIN_TV           = 12                                 ' Base TV Pin

  NONE             = 0                                  ' Wave shapes for VCO
  SINE             = 1
  TRIANGLE         = 2
  SQUARE           = 3
  SAW_UP           = 4
  SAW_DOWN         = 5
  RAMP_UP          = 6
  RAMP_DOWN        = 7
  SPIKE            = 8
  SPIKE_UP         = 9
  SPIKE_DOWN       = 10
  NOISE            = 11
  
  VCO_OUTPUTLEVEL  = 0 * 4                              ' WR - VCO control parameters
  VCO_RESET        = 1 * 4                              ' RD 
  VCO_FREQ         = 2 * 4                              ' WR
  VCO_FREQSTEPRATE = 3 * 4                              ' RD
  VCO_WAVETYPE     = 4 * 4                              ' WR
  VCO_TABLEINDEX   = 5 * 4                              ' RD
  VCO_AMPTR        = 6 * 4                              ' WR
  VCO_AMDEPTH      = 7 * 4                              ' WR

  VCO_PARAM_MAX    = 7 * 4                              ' WR = Write/Read, RD = Read Only

  VCO_PARAM_SIZE   = ( VCO_PARAM_MAX / 4 ) + 1
    
VAR

  long vco1[ VCO_PARAM_SIZE ]                           ' 1kHz tone generating VCO
  long vco2[ VCO_PARAM_SIZE ]                           ' Amplitude Modulator
  
PUB Synth | freq

  tv.start( PIN_TV )

  StartSynthPlayerInOwnCog

  long[ @vco1 + VCO_WAVETYPE ] := SINE                  ' 1kHz VCO modulated by VCO2
  long[ @vco1 + VCO_FREQ     ] := 1_000                 
  long[ @vco1 + VCO_AMPTR    ] := @vco2                 ' Modulate from VCO2 (LFO)
  long[ @vco1 + VCO_AMDEPTH  ] := BytePercent(75)       ' Modulation Depth = 75%, Output Level = 25%..100%

  long[ @vco2 + VCO_WAVETYPE ] := TRIANGLE              ' 1Hz Frequency Modulator
  long[ @vco2 + VCO_FREQ     ] := 1                     

  repeat                                                ' Runs all by itself
    tv.out($01)
    tv.hex(long[ @vco1 + VCO_FREQSTEPRATE ],8)
    tv.out(" ")
    tv.dec(long[ @vco1 + VCO_FREQ ])
    tv.str(String("Hz",$0D))
    tv.hex(long[ @vco2 + VCO_FREQSTEPRATE ],8)
    tv.out(" ")
    tv.dec(long[ @vco2 + VCO_FREQ ])
    tv.str(String("Hz "))
    tv.hex(long[ @vco2 + VCO_OUTPUTLEVEL ],8)           ' This is the VCO2 output modulation level

PRI BytePercent( percent )

  result := ( ( percent <# 100 ) * 255 ) / 100
  
PRI PauseMs( mS )

  waitcnt( CLKFREQ / 1000 * mS + CNT )
  
' ******************************************************************************************************

CON

  TIMEOUT_TICKS = 6_000

VAR

  long freqAtOneStepPerTick
  long noiseSource
  
  long stackSynthPlayer[50]                             ' Guessed size

PUB StartSynthPlayerInOwnCog

  cognew( SynthPlayer, @stackSynthPlayer )
  
  pauseMs(100) ' Give the SynthPlayer time to get loaded and running

PRI SynthPlayer

  if ( PIN_LEFT => 0 ) and ( PIN_LEFT =< 31 )
    DIRA[ PIN_LEFT  ] := 1

  if ( PIN_RIGHT => 0 ) and ( PIN_RIGHT =< 31 ) and ( PIN_LEFT <> PIN_RIGHT)
    DIRA[ PIN_RIGHT ] := 1

  CTRA := %00110 << 26 + 0 << 9 + PIN_LEFT
  CTRB := %00110 << 26 + 0 << 9 + PIN_RIGHT

  freqAtOneStepPerTick := CLKFREQ / ( TIMEOUT_TICKS * 256 )

  noiseSource := CNT
  
  repeat
    Vco( @vco1 )                                        ' The 1kHz VCO
    Vco( @vco2 )                                        ' The 1Hz modulator VCO (LFO)
    Dac( @vco1 )
    waitcnt( CNT + TIMEOUT_TICKS )                      ' Very poor quality timing loop
 
PRI Vco( vcoPtr ) | vcoTableIndex, vcoAmPtr, vcoAmMultiplier

  long[ vcoPtr + VCO_FREQSTEPRATE ] := $01_00_00_00 / freqAtOneStepPerTick * long[ vcoPtr + VCO_FREQ ]
 
  if long[ vcoPtr + VCO_RESET ]
    long[ vcoPtr + VCO_RESET ] := 0
    vcoTableIndex := 0
  else
    vcoTableIndex := long[ vcoPtr + VCO_TABLEINDEX ]
  
  case long[ vcoPtr + VCO_WAVETYPE ]
    SINE :
      result := ( byte[ @SineTable + ( vcoTableIndex >> 24 ) ] ) << 24
    TRIANGLE :               
      if vcoTableIndex => 0
        result := ( vcoTableIndex << 1 )
      else
        result := ( vcoTableIndex << 1 ) ^ $FF_FF_FF_FF
    SQUARE :
      result := vcoTableIndex < 0
    SAW_UP :
      result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) + $40 ) ] ^ $FF ) << 24
    SAW_DOWN :
      result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) + $40 ) ]       ) << 24
    RAMP_UP :
      result := vcoTableIndex
    RAMP_DOWN :
      result := vcoTableIndex ^ $FF_FF_FF_FF
    SPIKE :
      case vcoTableIndex >> 30
        %00 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 ) ^ $80 ) ] | $80 ) << 24
        %01 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 )       ) ] & $7F ) << 24
        %10 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 ) ^ $80 ) ] & $7F ) << 24
        %11 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 )       ) ] | $80 ) << 24
    SPIKE_UP :
      if vcoTableIndex => 0
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 )       ) ] ^ $80 ) << 24
      else
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) | $C0 ) ] ^ $80 ) << 24
    SPIKE_DOWN :
      if vcoTableIndex => 0
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 )       ) ] ^ $7F ) << 24
      else
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) | $C0 ) ] ^ $7F ) << 24
    NOISE :
      ? noiseSource                                       
      result := ( vcoTableIndex  & $80_00_00_00 ) | ( noiseSource & $7F_FF_FF_FF )
    other :
      result := $80_00_00_00

  long[ vcoPtr + VCO_TABLEINDEX ] := vcoTableIndex + long[ vcoPtr + VCO_FREQSTEPRATE ]

  ' Amplitude Modulation code below

  if ( vcoAmPtr := long [ vcoPtr + VCO_AMPTR ] )
    vcoAmMultiplier := 255 - ( ( ( ( ( long[ vcoAmPtr + VCO_OUTPUTLEVEL ] >> 24 ) ^ $FF ) * long[ vcoPtr + VCO_AMDEPTH ] ) >> 8 ))
    if result => 0
      result := ( ( ( result ^ $7F_FF_FF_FF ) >> 8 ) * vcoAmMultiplier ) ^ $7F_FF_FF_FF
    else
      result := ( ( ( result & $7F_FF_FF_FF ) >> 8 ) * vcoAmMultiplier ) | $80_00_00_00
  
  ' Amplitude Modulaton code above
  
  long[ vcoPtr + VCO_OUTPUTLEVEL ] := result
               
PRI Dac( vcoPtr )

  FRQA := long[ vcoPtr + VCO_OUTPUTLEVEL ]
  FRQB := FRQA

DAT

SineTable     byte 128,131,134,137,140,143,146,149,152,156,159,162,165,168,171,174
              byte 176,179,182,185,188,191,193,196,199,201,204,206,209,211,213,216
              byte 218,220,222,224,226,228,230,232,234,236,237,239,240,242,243,245
              byte 246,247,248,249,250,251,252,252,253,254,254,255,255,255,255,255
              byte 255,255,255,255,255,255,254,254,253,252,252,251,250,249,248,247
              byte 246,245,243,242,240,239,237,236,234,232,230,228,226,224,222,220
              byte 218,216,213,211,209,206,204,201,199,196,193,191,188,185,182,179
              byte 176,174,171,168,165,162,159,156,152,149,146,143,140,137,134,131
              byte 128,124,121,118,115,112,109,106,103, 99, 96, 93, 90, 87, 84, 81 
              byte  79, 76, 73, 70, 67, 64, 62, 59, 56, 54, 51, 49, 46, 44, 42, 39 
              byte  37, 35, 33, 31, 29, 27, 25, 23, 21, 19, 18, 16, 15, 13, 12, 10 
              byte   9,  8,  7,  6,  5,  4,  3,  3,  2,  1,  1,  0,  0,  0,  0,  0  
              byte   0,  0,  0,  0,  0,  0,  1,  1,  2,  3,  3,  4,  5,  6,  7,  8  
              byte   9, 10, 12, 13, 15, 16, 18, 19, 21, 23, 25, 27, 29, 31, 33, 35 
              byte  37, 39, 42, 44, 46, 49, 51, 54, 56, 59, 62, 64, 67, 70, 73, 76 
              byte  79, 81, 84, 87, 90, 93, 96, 99,103,106,109,112,115,118,121,124
          