{

AiChip_Synth_005.spin
=====================

For an overview description see the AiChip_Synth_001.spin file.
                                                          
This is an extension of AiChip_Synth_004.spin with ADSR Envelope Generator added.

The ADSR has four phases, Attack, Decay, Sustain and Release. Whenever ADSR_GATE is set to 1
the Attack phase is started and the envelope will increase to 100% from wherever it is. When it
reaches 100% the Decay phase is entered and the envelope will decrease to the ADSR_SUSTAINLEVEL
and stay there until the ADSR_GATE is set to 1 ( a new note ) or to 0 ( note released ). When
ADSR_GATE is set to 0 the Release phase is entered and the envelope will decrease from wherever
it is to zero.

Note that Attack and Release phases can be entered regardless of whatever phase the ADSR is in
so a note can be played staccato and the attack may not have completed before it is released.

Setting the ADSR_SUSTAINLEVEL to zero means the envelope will always Decay to zero once the
Attack phase completes regardless of however long a note is held for.

Setting ADSR_ADREPEAT to 1 will cause the envelope to repeat the Attack and Decay phases until
the note is released. This allows a Tremolo ( varying amplitude ) effect to be achieved while
a note is held on.
   
NOTE : ADSR_ISLOGARITHMIC is not currently used. UnsignedAdd() and UnsignedSub() are not well
implemented.

NOTE : Added RANDOM waveform. NOISE biases the noise source so it's below $80_00_00_00 in the
first cycle of the waveform, above in the second, RANDOM doesn't, it's all over the place.

NOTE : VCO_AMDEPTH has changed to using the MSB of a long instead of being an LSB byte so it
is compatible with other modulation sources. VCO parameters re-ordered to put read-only ( those
set or updated within Vco() parameters ) at the end of the list.

}

CON

  _CLKMODE                      = XTAL1 + PLL16x        ' 80MHz System Clock Speed
  _XINFREQ                      = 5_000_000

OBJ

  tv                            : "TV_Text"             ' Debugging only - can be deleted
  
CON

  PIN_LEFT                      = 10                    ' Set -Ve if not used
  PIN_RIGHT                     = 11                    ' Set -Ve if not used
                                                                                
  PIN_TV                        = 12                    ' Base TV Pin

  NONE                          = 0                     ' Wave shapes for VCO
  SINE                          = 1
  TRIANGLE                      = 2
  SQUARE                        = 3
  SAW_UP                        = 4
  SAW_DOWN                      = 5
  RAMP_UP                       = 6
  RAMP_DOWN                     = 7
  SPIKE                         = 8
  SPIKE_UP                      = 9
  SPIKE_DOWN                    = 10
  NOISE                         = 11
  RANDOM                        = 12
  
  VCO_OUTPUTLEVEL               = 0 * 4                 ' WR  VCO control parameters
  VCO_RESET                     = 1 * 4                 ' WR, 1 = Reset 
  VCO_FREQ                      = 2 * 4                 ' WR, Hz
  VCO_WAVETYPE                  = 3 * 4                 ' WR
  VCO_AMPTR                     = 4 * 4                 ' WR
  VCO_AMDEPTH                   = 5 * 4                 ' WR
  VCO_FREQSTEPRATE              = 6 * 4                 ' RD
  VCO_TABLEINDEX                = 7 * 4                 ' RD

  VCO_PARAM_MAX                 = 7 * 4                 ' WR = Write/Read, RD = Read Only

  VCO_PARAM_SIZE                = ( VCO_PARAM_MAX / 4 ) + 1

  PHASE_OFF                     = 0                     ' ADSR Phases
  PHASE_ATTACK                  = 1
  PHASE_DECAY                   = 2
  PHASE_SUSTAIN                 = 3
  PHASE_RELEASE                 = 4

  ADSR_OUTPUTLEVEL              = 0 * 4                 ' RD  ADSR control parameters
  ADSR_GATE                     = 1 * 4                 ' WR, 1 = Start, 0 = Stop
  ADSR_ATTACKTIME               = 2 * 4                 ' WR, mS
  ADSR_DECAYTIME                = 3 * 4                 ' WR, mS
  ADSR_RELEASETIME              = 4 * 4                 ' WR, mS
  ADSR_SUSTAINLEVEL             = 5 * 4                 ' WR
  ADSR_ADREPEAT                 = 6 * 4                 ' WR, 1 = Repeat Attack-Decay, 0 = Sustain
  ADSR_ISLOGARITHMIC            = 7 * 4                 ' WR, 1 = Log Attack-Decay-Release, 0 = Linear
  ADSR_ATTACKSTEPRATE           = 8 * 4                 ' RD
  ADSR_DECAYSTEPRATE            = 9 * 4                 ' RD
  ADSR_RELEASESTEPRATE          = 10 * 4                ' RD
  ADSR_RAMP                     = 11 * 4                ' RD
  ADSR_PHASE                    = 12 * 4                ' RD

  ADSR_PARAM_MAX                = 12 * 4

  ADSR_PARAM_SIZE               = ( ADSR_PARAM_MAX / 4 ) + 1
    
VAR

  long vco1[ VCO_PARAM_SIZE ]                           ' 1kHz tone generating VCO

  long adsr1[ ADSR_PARAM_SIZE ]                         ' ADSR Envelope Generator
  
PUB Synth | freq

  tv.start( PIN_TV )

  StartSynthPlayerInOwnCog

  long[ @vco1 + VCO_WAVETYPE ] := SINE                  ' 1kHz VCO modulated by ADSR1
  long[ @vco1 + VCO_FREQ     ] := 1_000                 
  long[ @vco1 + VCO_AMPTR    ] := @adsr1                ' Modulate from ADSR
  long[ @vco1 + VCO_AMDEPTH  ] := LongPercent(100)      ' Modulation Depth = 100%

  long[ @adsr1 + ADSR_ATTACKTIME   ] := 1_000           ' 1000ms / 1s so we can hear the effect
  long[ @adsr1 + ADSR_DECAYTIME    ] := 1_000
  long[ @adsr1 + ADSR_RELEASETIME  ] := 1_000
  long[ @adsr1 + ADSR_SUSTAINLEVEL ] := LongPercent(50)

  repeat
    NoteOn
    PauseMs(1000)
    NoteOff
    PauseMs(2000)

PRI NoteOn

    long[ @adsr1 + ADSR_GATE ] := 1
    repeat while long[ @adsr1 + ADSR_GATE ] == 1
    repeat until long[ @adsr1 + ADSR_PHASE ] == PHASE_SUSTAIN
      Report
    Report

PRI NoteOff

    long[ @adsr1 + ADSR_GATE ] := 0
    repeat until long[ @adsr1 + ADSR_PHASE ] == PHASE_OFF
      Report
    Report
    
PRI Report

  tv.out($01)
  tv.hex(long[ @adsr1 + ADSR_PHASE ], 1)
  tv.out(" ")
  tv.out(LookupZ(long[ @adsr1 + ADSR_PHASE ] : "-","A","D","S","R"))
  tv.out(" ")
  tv.hex(long[ @adsr1 + ADSR_OUTPUTLEVEL ], 8)
    
PRI LongPercent( percent )

  result := ( ( ( percent <# 100 ) * 255 ) / 100 ) << 24
  if ( result & $01_00_00_00 )
    result |= $00_FF_FF_FF
    
PRI PauseMs( mS )

  waitcnt( CLKFREQ / 1000 * mS + CNT )
  
' ******************************************************************************************************

CON

  TIMEOUT_TICKS = 8_000

VAR

  long freqHzAtOneStepPerTick
  long timeUsAtOneStepPerTick

  long freqStepRatePerHz
  long timeStepRatePerMs
  
  long noiseSource
  
  long stackSynthPlayer[50]                             ' Guessed size

PUB StartSynthPlayerInOwnCog

  cognew( SynthPlayer, @stackSynthPlayer )
  
  pauseMs(100) ' Give the SynthPlayer time to get loaded and running

PRI SynthPlayer | timeout

  if ( PIN_LEFT => 0 ) and ( PIN_LEFT =< 31 )
    DIRA[ PIN_LEFT  ] := 1

  if ( PIN_RIGHT => 0 ) and ( PIN_RIGHT =< 31 ) and ( PIN_LEFT <> PIN_RIGHT)
    DIRA[ PIN_RIGHT ] := 1

  CTRA := %00110 << 26 + 0 << 9 + PIN_LEFT
  CTRB := %00110 << 26 + 0 << 9 + PIN_RIGHT

  freqHzAtOneStepPerTick := CLKFREQ / ( TIMEOUT_TICKS * 256 )
  timeUsAtOneStepPerTick := TIMEOUT_TICKS / ( CLKFREQ / 1_000_000 )

  freqStepRatePerHz := ( $01_00_00_00 / freqHzAtOneStepPerTick ) << 1
  timeStepRatePerMs := ( $01_00_00_00 / timeUsAtOneStepPerTick ) << 1 / 1000
  
  noiseSource := CNT

  repeat
    Adsr( @adsr1 )                                      ' The ADSR Envelope Generator
    Vco( @vco1 )                                        ' The 1kHz VCO
    Dac( @vco1 )
    waitcnt( CNT + TIMEOUT_TICKS )                      ' Very poor quality timing loop

PRI Adsr( adsrPtr ) | level, stepRate

  long[ adsrPtr + ADSR_ATTACKSTEPRATE  ] := timeStepRatePerMs * long[ adsrPtr + ADSR_ATTACKTIME  ]
  long[ adsrPtr + ADSR_DECAYSTEPRATE   ] := timeStepRatePerMs * long[ adsrPtr + ADSR_DECAYTIME   ]
  long[ adsrPtr + ADSR_RELEASESTEPRATE ] := timeStepRatePerMs * long[ adsrPtr + ADSR_RELEASETIME ]

  case long[ adsrPtr + ADSR_GATE ]
    0 : long[ adsrPtr + ADSR_PHASE ] := PHASE_RELEASE
    1 : long[ adsrPtr + ADSR_GATE ] := 2
        long[ adsrPtr + ADSR_PHASE ] := PHASE_ATTACK

  level := long[ adsrPtr + ADSR_OUTPUTLEVEL]

  case long[ adsrPtr + ADSR_PHASE ]
    PHASE_ATTACK :
      stepRate := long[ adsrPtr + ADSR_ATTACKSTEPRATE ]
      level := UnsignedAdd( level, steprate )
      if ( level >> 24 ) == $FF
        level := $FFFF_FFFF
        long[ adsrPtr + ADSR_PHASE ] := PHASE_DECAY
    PHASE_DECAY :
      stepRate := long[ adsrPtr + ADSR_DECAYSTEPRATE ]
      level := UnsignedSub( level, steprate )
      if ( level >> 24 ) == ( long[ adsrPtr + ADSR_SUSTAINLEVEL ] >> 24 )
        level := long[ adsrPtr + ADSR_SUSTAINLEVEL ]
        if long[ adsrPtr + ADSR_ADREPEAT]
          long[ adsrPtr + ADSR_PHASE ] := PHASE_ATTACK
        else
          long[ adsrPtr + ADSR_PHASE ] := PHASE_SUSTAIN
    PHASE_RELEASE :
      stepRate := long[ adsrPtr + ADSR_RELEASESTEPRATE ]
      level := UnsignedSub( level, steprate )
      if ( level >> 24 ) == 0
        level := 0
        long[ adsrPtr + ADSR_PHASE ] := PHASE_OFF

  long[ adsrPtr + ADSR_OUTPUTLEVEL ] := level
   
PRI Vco( vcoPtr ) | vcoTableIndex, vcoAmPtr, vcoAmMultiplier

  long[ vcoPtr + VCO_FREQSTEPRATE ] := freqStepRatePerHz * long[ vcoPtr + VCO_FREQ ]
 
  if long[ vcoPtr + VCO_RESET ]
    long[ vcoPtr + VCO_RESET ] := 0
    vcoTableIndex := 0
  else
    vcoTableIndex := long[ vcoPtr + VCO_TABLEINDEX ]
  
  case long[ vcoPtr + VCO_WAVETYPE ]
    SINE :
      result := ( byte[ @SineTable + ( vcoTableIndex >> 24 ) ] ) << 24
    TRIANGLE :               
      if vcoTableIndex => 0
        result := ( vcoTableIndex << 1 )
      else
        result := ( vcoTableIndex << 1 ) ^ $FF_FF_FF_FF
    SQUARE :
      result := vcoTableIndex < 0
    SAW_UP :
      result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) + $40 ) ] ^ $FF ) << 24
    SAW_DOWN :
      result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) + $40 ) ]       ) << 24
    RAMP_UP :
      result := vcoTableIndex
    RAMP_DOWN :
      result := vcoTableIndex ^ $FF_FF_FF_FF
    SPIKE :
      case vcoTableIndex >> 30
        %00 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 ) ^ $80 ) ] | $80 ) << 24
        %01 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 )       ) ] & $7F ) << 24
        %10 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 ) ^ $80 ) ] & $7F ) << 24
        %11 : result := ( byte[ @SineTable + ( ( vcoTableIndex >> 24 )       ) ] | $80 ) << 24
    SPIKE_UP :
      if vcoTableIndex => 0
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 )       ) ] ^ $80 ) << 24
      else
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) | $C0 ) ] ^ $80 ) << 24
    SPIKE_DOWN :
      if vcoTableIndex => 0
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 )       ) ] ^ $7F ) << 24
      else
        result := ( byte[ @SineTable + ( ( vcoTableIndex >> 25 ) | $C0 ) ] ^ $7F ) << 24
    NOISE :
      ? noiseSource                                       
      result := ( vcoTableIndex  & $80_00_00_00 ) | ( noiseSource & $7F_FF_FF_FF )
    RANDOM :
      ? noiseSource                                       
      result := noiseSource
    other :
      result := $80_00_00_00

  long[ vcoPtr + VCO_TABLEINDEX ] := vcoTableIndex + long[ vcoPtr + VCO_FREQSTEPRATE ]

  if ( vcoAmPtr := long [ vcoPtr + VCO_AMPTR ] )
    vcoAmMultiplier := 255 - ( ( ( ( ( long[ vcoAmPtr + VCO_OUTPUTLEVEL ] >> 24 ) ^ $FF ) * ( long[ vcoPtr + VCO_AMDEPTH ] >> 24 ) ) >> 8 ))
    if result => 0
      result := ( ( ( result ^ $7F_FF_FF_FF ) >> 8 ) * vcoAmMultiplier ) ^ $7F_FF_FF_FF
    else
      result := ( ( ( result & $7F_FF_FF_FF ) >> 8 ) * vcoAmMultiplier ) | $80_00_00_00
  
  long[ vcoPtr + VCO_OUTPUTLEVEL ] := result

PRI Dac( vcoPtr )

  FRQA := long[ vcoPtr + VCO_OUTPUTLEVEL ]
  FRQB := FRQA

PRI UnsignedAdd( lhs, rhs )

  ' **** This is a hack and doesn't work for all cases

  result := ( lhs >> 1 ) + ( rhs >> 1 )
  if result & $80_00_00_00
    result := $FF_FF_FF_FF
  else
    result <<= 1

PRI UnsignedSub( lhs, rhs )

  ' **** This is a hack and doesn't work for all cases

  result := ( lhs >> 1 ) - ( rhs >> 1 )
  if result & $80_00_00_00
    result := $00_00_00_00
  else
    result <<= 1

DAT

SineTable     byte 128,131,134,137,140,143,146,149,152,156,159,162,165,168,171,174
              byte 176,179,182,185,188,191,193,196,199,201,204,206,209,211,213,216
              byte 218,220,222,224,226,228,230,232,234,236,237,239,240,242,243,245
              byte 246,247,248,249,250,251,252,252,253,254,254,255,255,255,255,255
              byte 255,255,255,255,255,255,254,254,253,252,252,251,250,249,248,247
              byte 246,245,243,242,240,239,237,236,234,232,230,228,226,224,222,220
              byte 218,216,213,211,209,206,204,201,199,196,193,191,188,185,182,179
              byte 176,174,171,168,165,162,159,156,152,149,146,143,140,137,134,131
              byte 128,124,121,118,115,112,109,106,103, 99, 96, 93, 90, 87, 84, 81 
              byte  79, 76, 73, 70, 67, 64, 62, 59, 56, 54, 51, 49, 46, 44, 42, 39 
              byte  37, 35, 33, 31, 29, 27, 25, 23, 21, 19, 18, 16, 15, 13, 12, 10 
              byte   9,  8,  7,  6,  5,  4,  3,  3,  2,  1,  1,  0,  0,  0,  0,  0  
              byte   0,  0,  0,  0,  0,  0,  1,  1,  2,  3,  3,  4,  5,  6,  7,  8  
              byte   9, 10, 12, 13, 15, 16, 18, 19, 21, 23, 25, 27, 29, 31, 33, 35 
              byte  37, 39, 42, 44, 46, 49, 51, 54, 56, 59, 62, 64, 67, 70, 73, 76 
              byte  79, 81, 84, 87, 90, 93, 96, 99,103,106,109,112,115,118,121,124
          